
# THIS FILE IS GENERATED FROM CVXPY SETUP.PY
short_version = '1.7.5'
version = '1.7.5'
full_version = '1.7.5'
git_revision = '68141a6'
commit_count = '68141a6'
release = True
if not release:
    version = full_version
